#!/usr/bin/env python
"""GenBank - reference human genome assembly."""

from ..resource import BaseResource
from ..servers.ncbi import NCBI
import sh


class Resource(BaseResource):
  """docstring for CCDS Resource"""
  def __init__(self):
    super(Resource, self).__init__()

    self.id = "genbank"

    self.ftp = NCBI()
    self.baseUrl = ("genbank/genomes/Eukaryotes/vertebrates_mammals/"
                    "Homo_sapiens")

    self.parts = 24
    self.names = ["Genbank.Homo_sapiens.{chrom}.fa.gz".format(chrom=chrom)
                  for chrom in range(1, 23) + ["X", "Y"]]

  def versions(self):
    return [dirName for dirName in self.ftp.listFiles(self.baseUrl, "GRCh*")]

  def latest(self):
    # Extract assembly and patch number
    versions = [(int(v[4:6]), int(v[8:] or 0)) for v in self.versions()]

    # Sort based on assembly and patch => return the last item
    combo = sorted(versions, key=lambda x: (x[0], x[1]))[-1]

    # Compose version string
    version = "GRCh{v}".format(v=combo[0], patch=combo[1])

    # Only add patch extention for numbers > 0
    if combo[1] != 0:
      version += ".p" + combo[1]

    return version

  def newer(self, current, challenger):
    # Extract assembly and patch number
    combos = [(int(v[4:6]), int(v[8:] or 0)) for v in (current, challenger)]

    # If the assemblies differ
    if combos[0][0] != combos[1][0]:
      return combos[1][0] > combos[0][0]

    else:
      # The assemblies are the same => compare patch numbers
      return combos[1][1] > combos[0][1]

  def paths(self, version):
    base = ("{base}/{v}/Primary_Assembly/assembled_chromosomes/FASTA"
            .format(base=self.baseUrl, v=version))
    files = self.ftp.listFiles(base, "*.fa.gz")

    # Should match each of the 24 chromosomes (not MT)
    return ["{base}/{file}".format(base=base, file=f) for f in files]

  def postClone(self, cloned_files, target_dir, version):
    """
    .. versionadded:: 0.3.0
    """
    # Start by extracting all the files
    for f in cloned_files:
      # GunZIP the file (and remove the archive)
      sh.gunzip(f)

    # Then let's concat them
    target_path = "{}/Genbank.Homo_sapiens.fa".format(target_dir)
    # Remove ".gz" ending to point to extracted files
    cat_args = [f[:-3] for f in cloned_files]

    # Execute the concatenation in the background and write to the target path
    sh.cat(*cat_args, _out=target_path)



#!/usr/bin/env python
"""
A genomics resource.

Each resource is responsible for:
- Fetching a list of availble versions
- Figuring out which the latest version is
- Comparing and determining which of two versions are the latest
- Returning link(s) for the files to download
- Specify binary/gzip/concat options
- Keeps track of what the resource is called locally
"""


class BaseResource(object):
  """
  A Resource represents a local genomics resource that can be on the file
  system or destined to be downloaded.
  """
  def __init__(self):
    super(BaseResource, self).__init__()
    # The server for the resource
    self.server = None

    # The keyword for the resource
    self.id = "resource"

  def versions(self):
    """
    <public> Returns a list of version tags for availble resource versions.

    :returns: List of supported resource releases
    :rtype: list
    """
    return []

  def latest(self):
    """
    <public> Returns the version tag for the latest availble version of the
    resource.

    :returns: The most up-to-date release tag
    """
    return ""

  def newer(self, current, challenger):
    """
    <public> Compares two different version tags and returns ``True`` if the
    `challenger` is newer than the `current` tag.

    :param str current: Version tag to compare against
    :param str challenger: Version tag to compare with
    :returns: ANS: Challenger is newer than current
    :rtype: bool
    """
    return current > challenger

  def paths(self, version):
    """
    <public> Returns the full download path matching the given ``version`` tag.

    :param str version: Version tag of the release
    :returns: List of full paths for included files
    :rtype: list
    """
    return []

  def postClone(self, cloned_files, target_dir, version):
    """
    <public> This callback method will be called once the files in the resource
    have been successfully downloaded. The paths to the files will be provided
    as a list to the method.

    This can be used as a way to rename, unzip, or concat files; generally
    post process them to prepare them for the user.

    :param list cloned_files: List of paths to the downloaded resource files
    :param str target_dir: Path to resource directory
    :param object version: Version of the resource that was downloaded

    .. versionadded:: 0.3.0
    """
    return 0



#!/usr/bin/env python
"""UCSC Human genome assembly."""

from __future__ import print_function
from ..resource import BaseResource
from ..servers.ucsc import UCSC
import sh


class Resource(BaseResource):
  """docstring for Ensembl Assembly Resource"""
  def __init__(self):
    super(Resource, self).__init__()

    self.id = "ucsc_assembly"

    self.ftp = UCSC()
    self.baseUrl = "goldenPath"

    self.parts = 1
    self.names = ["UCSC.Homo_sapiens.tar.gz"]

  def versions(self):
    # Only one release
    # Skip releases like "hg15june2000" (if)
    return [dirName for dirName in
            self.ftp.listFiles(self.baseUrl, "hg[1-9]*")
            if not len(dirName) > 4]

  def latest(self):
    number = max([int(v.replace("hg", "")) for v in self.versions()])
    return "hg{}".format(number)

  def newer(self, current, challenger):
    return int(challenger.replace("hg", "")) > int(current.replace("hg", ""))

  def paths(self, version):
    base = "{base}/{v}/bigZips".format(base=self.baseUrl, v=version)

    # If the version is newer than 'hg18' it will be a '.tar.gz' file
    if self.newer("hg18", version):
      return ["{base}/chromFa.tar.gz".format(base=self.baseUrl)]

    else:
      return ["{base}/chromFa.zip".format(base=base)]

  def postClone(self, cloned_files, target_dir, version):
    """
    Extracts the compressed archives.

    .. versionadded:: 0.3.0
    """
    f = cloned_files[0]

    if self.newer("hg18", version):
      # GZIP and TAR the file and save to the target directory
      sh.tar("-xzf", f, "-C", target_dir)

    else:
      # Rename to ".zip"
      sh.mv(f, f.replace("tar.gz", "zip"))

      # GunZIP the file (and remove the archive)
      sh.gunzip(f)



#!/usr/bin/env python
"""1000 Genomes High Confidence SNPs."""

# We can base this on another GATK resource since it's pretty much the same
from example import Resource as iResource


class Resource(iResource):
  """docstring for 1000 Genomes SNPs Resource"""
  def __init__(self):
    super(Resource, self).__init__()

    self.id = "1000g_snps"

    self.parts = 1
    self.names = ['1000G_phase1.snps.high_confidence.vcf.gz']

  def paths(self, version):
    # 1 file
    bundle_id, assembly = self.defineVersion(version)
    base = "{base}/{bundle}/{assembly}".format(base=self.baseUrl,
                                               bundle=bundle_id,
                                               assembly=assembly)
    f = "1000G_phase1.snps.high_confidence.{}.vcf.gz".format(assembly)

    return ["{base}/{file}".format(base=base, file=f)]



#!/usr/bin/env python
"""dbSNP - genetic variation within and across different species."""

# We can use another GATK resource since it's pretty much the same
from example import Resource as iResource


class Resource(iResource):
  """docstring for dbSNP Resource"""
  def __init__(self):
    super(Resource, self).__init__()

    self.id = "dbsnp"

    self.parts = 1
    self.names = ["dbsnp.vcf.gz"]

  def paths(self, version):
    bundle_id, assembly = self.defineVersion(version)

    if bundle_id == self.latest():
      num = "138"
    else:
      num = "137"

    # 1 file
    base = "{base}/{bundle}/{assembly}".format(base=self.baseUrl,
                                               bundle=bundle_id,
                                               assembly=assembly)
    f = "dbsnp_{num}.{assembly}.vcf.gz".format(num=num, assembly=assembly)

    return ["{base}/{file}".format(base=base, file=f)]



#!/usr/bin/env python
"""The Consensus CoDing Sequence (CCDS) project; "a core set of human and mouse protein-coding regions"."""

from ..resource import BaseResource
from ..servers.ncbi import NCBI


class Resource(BaseResource):
  """docstring for CCDS Resource"""
  def __init__(self):
    super(Resource, self).__init__()

    self.id = "ccds"

    self.ftp = NCBI()
    self.baseUrl = "pub/CCDS"

    self.parts = 1
    self.names = ["CCDS.txt"]

  def versions(self):
    return [dirName for dirName in self.ftp.ls(self.baseUrl + "/archive")
            if dirName.startswith("Hs")]

  def latest(self):
    # Load in info-file
    f = self.ftp.file("{}/current_human/BuildInfo.current.txt"
                      .format(self.baseUrl))

    # Discard comment line
    _ = f.readline()

    # Return formatted NCBI release number
    return "Hs{}".format(f.readline().split("\t")[1])

  def newer(self, current, challenger):
    currentFloat = float(current.replace("Hs", ""))
    challengerFloat = float(challenger.replace("Hs", ""))

    # Simply check which float is biggest
    return challengerFloat > currentFloat

  def paths(self, version):
    base = "{base}/archive/{v}".format(base=self.baseUrl, v=version)
    files = self.ftp.listFiles(base, "CCDS.[0-9]*")

    # Should only be one file that matches
    return ["{base}/{file}".format(base=base, file=f) for f in files]



#!/usr/bin/env python
from ..core import FTP


class GATK(FTP):
  """docstring for NCBI"""
  def __init__(self):
    super(GATK, self).__init__("ftp.broadinstitute.org",
                               "gsapubftp-anonymous", "")



#!/usr/bin/env python

import yaml
from path import path


class DefaultReader(object):
  """
  A default YAML reader. Allows for safe parsing, basic adding and fetching
  items as well as saving an updated YAML datastructre (``dict``).

  :param str yaml_path: (optional) Path to the YAML file to parse/save to.
  """
  def __init__(self, yaml_path=None):
    super(DefaultReader, self).__init__()
    self.items = {}

    # Load in the resources listed in the YAML file if provided
    if yaml_path:
      self.load(yaml_path)

  def __iter__(self):
    """
    <magic> Shortcut to iterate over all the keys and items.

    :returns: An iterable returning tuples: (key, item)
    """
    return iter(self.items.iteritems())

  def load(self, yaml_path=None):
    """
    <public> Parse a local YAML-formatted file. Raises an ``IOError`` if the
    path doesn't points to an existing file.

    .. code-block:: python

      >>> reader.load("~/Desktop/config.yaml")

    :param str yaml_path: (optional) A path to the YAML file to parse
    :returns: self
    """
    if yaml_path:
      self.source = path(yaml_path)

    if self.source.isfile():
      with open(self.source, "r") as handle:
        # Use safe load to protect from accidentally running malicious
        # functions defined in the YAML file. Limit to simple str/int values.
        self.items = yaml.safe_load(handle.read())

    else:
      # Initialize empty
      self.items = {}

    return self

  def add(self, key, item):
    """
    <public> Adds an item associated with a key to the bottom level of the
    YAML structure. The item could be any Python object but should be limited
    to simple strings, ints, or dicts of such objects.

    .. code-block:: python

      >>> reader.add("ccds", "Hs104")

    :param object key: Usually a string for use as an association key
    :param object item: Associated ``str``, ``int``, or ``dict``
    :returns: self
    """
    self.items[key] = item

    return self

  def save(self):
    """
    <public> Overwrites the file currently pointed to by the ``path``
    attribute with valid YAML formatting.

    :returns: self
    """
    with open(self.source, "w") as handle:
      handle.write(yaml.safe_dump(self.items, default_flow_style=False))

    return self

  def find(self, key=None, default=None):
    """
    <public> Returns all items or a single item associated with a given key.
    If a given key is not present the `default` object will be returned
    instead.

    .. code-block:: python

      >>> reader.find("ccds", default="latest")

    :param object key: (optional) Usually a string associated with the item
    :param object default: (optional) What to return if the key isn't found
    :returns: Either all items as a ``dict`` or a one matching the key
    """
    if key is None:
      # Return all
      return self.items
    else:
      return self.items.get(key, default)


class HistoryReader(DefaultReader):
  """
  A specialized YAML reader for dealing with Cosmid History files. Inherits
  from ``DefaultReader``.

  :param str yaml_path: Path to the cosmid history file (.cosmid.yaml)
  """
  def __init__(self, yaml_path):
    super(HistoryReader, self).__init__(yaml_path)

  def updateable(self, force=False):
    """
    <public> Returns all resources that could be updated.

    .. code-block:: python

      >>> history.updateable()
      {"ccds": "latest", ...}

    :param bool force: (optional) Include all resources by default
    :returns: A dict with ``{ resource_id: version_tag }``
    :rtype: dict
    """
    resources = {}
    for key, item in self.items.iteritems():

      # Exclude perfect target<->version matches
      # Always include resources with target=latest
      # If force is selected we include all resources
      if item['target'] != item['version'] or force:
        resources[key] = item['target']

    return resources


class ConfigReader(DefaultReader):
  """
  A specialized YAML reader for dealing with Cosmid History files. Inherits
  from ``DefaultReader``.

  :param str yaml_path: Path to the cosmid history file (.cosmid.yaml)
  """
  def __init__(self, yaml_path):
    super(ConfigReader, self).__init__(yaml_path)

  def addResource(self, resource_id, v_target="latest"):
    """
    <public> Adds a resources and version target to the config file.

    :param str resource_id: A valid resource ID
    :param str v_target: (optional) A version target to for the resource
    :returns: self
    """
    resources = self.find("resources", default=None)

    if resources is None:
      # Add the "resources" key first
      self.add("resources", {})

      # Recursively call the function
      self.addResource(resource_id, v_target)

    else:
      # We're good to go!
      resources[resource_id] = v_target

    return self

  def getResource(self, resource_id, default=None):
    """
    <public> Fetches a resource from the config file or returns `default`.

    :param str resource_id: A valid resource ID
    :param object default: (optional) What to return unless the ID matches
    :returns: A dict with a `resource_id`/`target` pair
    """
    resources = self.find("resources", default=None)

    if resources is None:
      return default

    else:
      return resources.get(resource_id, default)



#!/usr/bin/env python
"""Human genes from Ensembl."""

from ensembl_assembly import Resource as iResource
import sh


class Resource(iResource):
  """docstring for Ensembl GTF Resource"""
  def __init__(self):
    super(Resource, self).__init__()

    self.id = "gtf"

    self.names = ["Ensembl.Homo_sapiens.gtf.gz"]

  def paths(self, version):
    # 1 file
    base = "pub/release-{}/gtf/homo_sapiens".format(version)
    files = self.ftp.listFiles(base, "*gtf.gz")

    return ["{base}/{file}".format(base=base, file=files[0])]

  def postClone(self, cloned_files, target_dir, version):
    """
    Extracts the compressed archives.

    .. versionadded:: 0.3.0
    """
    # GZIP the files (and remove the archive)
    for f in cloned_files:

      # Only some of the files needs to be gzipped
      if f.endswith(".gz"):
        sh.gunzip(f)



#!/usr/bin/env python
"""1000 Genomes decoy assembly sequence, version 5."""

from ..resource import BaseResource
from ..servers.thousandg import ThousandG


class Resource(BaseResource):
  """docstring for CCDS Resource"""
  def __init__(self):
    super(Resource, self).__init__()

    self.id = "decoy"

    self.ftp = ThousandG()
    self.baseUrl = ("1000genomes/ftp/technical/reference/"
                    "phase2_reference_assembly_sequence")

    self.parts = 1
    self.names = ["hs37d5.fa.gz"]

  def versions(self):
    return [5]

  def latest(self):
    return 5

  def newer(self, current, challenger):
    return challenger > current

  def paths(self, version):
    uri = "{base}/hs37d{v}.fa.gz".format(base=self.baseUrl, v=version)
    return [uri]



#!/usr/bin/env python
"""Hapmap 3.3 from GATK."""

# We can another GATK resource since it's pretty much the same
from example import Resource as iResource


class Resource(iResource):
  """docstring for dbSNP Resource"""
  def __init__(self):
    super(Resource, self).__init__()

    self.id = "hapmap"

    self.parts = 1
    self.names = ["hapmap_3.3.vcf.gz"]

  def paths(self, version):
    bundle_id, assembly = self.defineVersion(version)

    # 1 file
    base = "{base}/{bundle}/{assembly}"\
           .format(base=self.baseUrl, bundle=bundle_id, assembly=assembly)
    f = "hapmap_3.3.{}.vcf.gz".format(assembly)

    return ["{base}/{file}".format(base=base, file=f)]



#!/usr/bin/env python
"""Ensembl - automatically annotated human genome reference."""

from ..resource import BaseResource
from ..servers.ensembl import Ensembl
import sh


class Resource(BaseResource):
  """docstring for Ensembl Assembly Resource"""
  def __init__(self):
    super(Resource, self).__init__()

    self.id = "ensembl_assembly"

    self.ftp = Ensembl()

    self.parts = 1
    self.names = ["Ensembl.Homo_sapiens.fa.gz"]

  def versions(self):
    releases = [int(directory.replace("release-", ""))
                for directory in self.ftp.listFiles("pub", "release-*")]

    # Releases before 46 have a different folder structure.
    return [num for num in releases if num > 46]

  def latest(self):
    return max(self.versions())

  def newer(self, current, challenger):
    # Simply check which int is bigger
    return challenger > current

  def paths(self, version):
    # 1 file
    base = "pub/release-{}/fasta/homo_sapiens/dna".format(version)
    files = self.ftp.listFiles(base, "*.dna.primary_assembly.fa.gz")

    return ["{base}/{file}".format(base=base, file=files[0])]

  def postClone(self, cloned_files, target_dir, version):
    """
    Extracts the downloaded assembly.

    .. versionadded:: 0.3.0
    """
    # GZIP and TAR the file and save to the target directory
    for f in cloned_files:
      sh.gunzip(f)

    return 0



#!/usr/bin/env python

from __future__ import print_function
from __future__ import division

import ftplib
from fnmatch import fnmatch
from StringIO import StringIO
import pkgutil
import importlib
from path import path
from fuzzywuzzy import process

import resources
from magicmethods import load_class
from yml import ConfigReader, HistoryReader
from messenger import Messenger


class FTP(object):
  """
  Model of a basic FTP server. Inherits a few methods from class:`ftplib.FTP`
  as well as extends with a few new methods making it more like `ftputil`.

  .. code-block::

    >>> from cosmid.core import FTP
    >>> ftp = FTP("ftp.ensembl.org", "anonymous", "")
    >>> ftp.ls("")
    ['ls-lR.gz',
     '.message',
     '.ok_to_rsync',
     'pub',
     ...
    ]

  :param str url: URL for the server to connect to
  :param str username: Username for an account on the server
  :param str password: Password to the accound on the server
  """
  def __init__(self, url, username, password):
    super(FTP, self).__init__()
    self.url = url
    self.username = username
    self.password = password

    # Connect to the FTP server
    self.ftp = ftplib.FTP(url, username, password)

    # Shortcuts
    self.nlst = self.ftp.nlst
    self.retrbinary = self.ftp.retrbinary
    self.sendcmd = self.ftp.sendcmd
    self.size = self.ftp.size

  def ls(self, dir_path="."):
    """
    <public> Functions like `ls` in Unix where it lists the folders and files
    in a specific directory. Compared to `nlst` it doesn't return the full
    path for each file/folder.

    :param str dir_path: (optional) Path to directory
    :returns: List of files/folders in the directory
    :rtype: list
    """
    return [path_.split("/")[-1] for path_ in self.nlst(dir_path)]

  def file(self, path):
    """
    <public> Open a file-like object for reading txt-files on the server
    without downloading it locally first.

    :param str path: Path to file
    :returns: File-like object
    :rtype: StringIO object
    """
    r = StringIO()

    self.retrbinary("RETR " + path, r.write)

    # Rewind the "file"
    r.seek(0)

    return r

  def fileSize(self, path):
    """
    <public> Returns the file size of a certain file on the server in MB.

    :param str path: Path to file
    :returns: Size of file in megabytes
    :rtype: int
    """
    # Switch to Binary mode (to be able to get size)
    self.sendcmd("TYPE i")

    return round(self.size(path)/1000000, 2)

  def listFiles(self, dirPath, pattern):
    """
    <public> Like `ls` but has the option to match file/folder names to a
    pattern.

    :param str dirPath: Path to directory
    :param str pattern: Glob-like pattern to match against files
    :returns: List of files/folders in the directory matching the pattern
    :rtype: list
    """
    return [item for item in self.ls(dirPath) if fnmatch(item, pattern)]

  def commit(self, fullPath, dest, mode=None):
    """
    <public>: Saves a file from the server, locally in the `dest`.

    :param str fullPath: Path from the cwd to the file to download
    :param str dest: Local path+filename where you want to save the file
    :param str mode: (optional) "b" for binary files
    :returns: 0: OK, >0: NOT OK
    """
    # Is the remote file gzipped? (Binary format)
    if mode is None:
      if fullPath.endswith(".gz") or fullPath.endswith(".bam"):
        mode = "b"
      else:
        # Default mode is to download non-binary files
        mode = ""

    # Open connection to the destination file and retrive the file
    with open(dest, "w" + mode) as handle:
      self.retrbinary("RETR " + fullPath, handle.write)


class Registry(object):
  """
  Hub of-sorts to talk with different `Cosmid` related files and resources. Can
  be seen as the API endpoint for `Cosmid`.
  """
  def __init__(self):
    super(Registry, self).__init__()

    # Set up YAML parser for optional config file
    self.config_path = path("cosmid.yaml")
    self.config = ConfigReader(self.config_path)

    # Extract stuff from config
    self.email = self.config.find("email")

    # Path to resource storage directory
    self.directory = path(self.config.find("directory", default="resources"))

    # Load history file consisting of already downloaded resources
    self.history_path = path(self.directory + "/.cosmid.yaml")
    self.history = HistoryReader(self.history_path)

    # Set up a :class:`cosmid.messenger.Messenger`
    self.messenger = Messenger("cosmid")

  def get(self, resource_id, type_="class"):
    """
    <public> Returns an instance of the specified resource class. Dodges an
    ``ImportError`` when failing to import a resource and returns ``None``
    instead.

    .. code-block:: python

      >>> resource = registry.get("ccds")
      >>> resource.latest()
      'Hs104'

    :param str resource_id: The resource key (name of module)
    :returns: A class instance of the resource
    """
    try:

      if type_ == "class":
        return load_class("cosmid.resources.{}.Resource".format(resource_id))()

      elif type_ == "module":
        return importlib.import_module("cosmid.resources." + resource_id)

      else:
          raise ValueError("Argument must be either 'class' or 'module'.")

    except ImportError:
      return None

  def grab(self, resource_id, target, collapse=False):
    """
    <public> Returns all that's nessesary to download a specific resource.
    The method will try to correct both ``resource_id`` and the ``target``
    release tag.

    :param str resource_id: What resource to download
    :param str target: What release of the resource to download
    """
    # Either import resource class or print warning and move on.
    # Test matching the resource ID
    options = [item[0] for item in self.ls()]
    resource_id = self.matchOne(resource_id, options)

    if resource_id is None:

      message = "Couldn't match resource ID: '{}'".format(resource_id)
      self.messenger.send("warning", message)

      return None, None, None, None

    # Get the resource
    resource = self.get(resource_id)

    # Now let's figure out the version
    # No specified version will match to the latest resource release
    if target == "latest":
      version = resource.latest()
    else:
      options = resource.versions()
      version = self.matchOne(target, options)

    if version is None:
      message = ("Couldn't match version '{id}#{v}'; {vers}"
                 .format(v=target, id=resource.id, vers=", ".join(options)))

      self.messenger.send("warning", message)

      return None, None, None, None

    # Get the goahead! (that we haven't already downloaded it)
    if self.goahead(resource, version):

      # Finally we can determine the paths to download and save the files
      dl_paths = resource.paths(version)

      if collapse:
        # The user can select to store all downloaded files in the same
        # directory
        resource_dir = ""

      else:
        # Or by default separate different resources into subdirectories
        resource_dir = "/" + resource.id

      save_paths = [("{dir}{mid}/{file}"
                     .format(dir=self.directory, mid=resource_dir, file=name))
                    for name in resource.names]

      # Add the resource to the history file as downloaded
      self.history.add(resource_id, {
        "version": version,
        "target": target,
        "names": resource.names,
        "sources": dl_paths
      })

      return resource, dl_paths, save_paths, version

    else:

      # The resource was already downloaded
      return None, None, None, None

  def ls(self):
    """
    <public> Returns a list of resource IDs and docstrings for all the
    included resource modules.

    *Reference*: http://stackoverflow.com/questions/1707709/list-all-the-modules-that-are-part-of-a-python-package

    .. code-block:: python

      >>> registry.ls()
      [('ccds', 'A curated database of generic element'), ...]

    :returns: A list of tuples: ``(resource_id, docstring)``
    :rtype: list
    """
    # Store everything here
    items = []

    prefix = resources.__name__ + "."
    # Fetch all the resource modules
    modules = pkgutil.iter_modules(resources.__path__, prefix)

    # Loop over all resource modules
    for importer, modpath, ispkg in modules:
      # Strip path
      modname = modpath.split(".")[-1]

      # Load the `Resource` class for the module
      module = self.get(modname, type_="module")

      # Save name and docstring
      items.append((modname, module.__doc__))

    return items

  def search(self, query, limit=5):
    """
    <public> Fuzzy matches a query string against each of the resource IDs and
    returns a limited number of results in order of match score.

    .. code-block:: python

      >>> registry.search("asmebly", limit=2)
      [('ensembl_assembly', 68),
       ('ncbi_assembly', 68)]

    :param str query: A string to match against the resource IDs
    :param int limit: (optional) A maximum number of results to return
    :returns: A list of tuples: ``(resource_id, score)`
    :rtype: list
    """
    # List all the available resources
    resources = self.ls()

    # Focus on resource IDs
    resource_ids = [resource[0] for resource in resources]

    # Fuzzy match against the resource IDs and return in order of best match
    return process.extract(query, resource_ids, limit=limit)

  def matchOne(self, target, options, threshold=60):
    """
    <public> Fuzzy matches e.g. a target version tag against a list of options.
    Returns the most likely match if the match score is sufficient.

    .. code-block:: python

      >>> resource = registry.get("ccds")
      >>> registry.matchOne(104, resource.versions())
      'Hs104'

      >>> registry.matchOne("ensembl", registry.ls())
      'ensembl_assembly'

    :param object target: Any Python object to match with
    :param list options: A list of possible options to match against
    :param int threshold: A lower threshold for accepting a best match
    :returns: The object with the best match (unless score is below threshold)
    :rtype: Python object
    """
    # Match against the options and extract the top match only
    result, score = process.extractOne(target, map(str, options))

    # Arbitrary lower limit for returning a *mathcing* result
    if score >= threshold:
      return result
    else:
      return None

  def goahead(self, resource, version):
    """
    Determines whether it's any idea in going ahead with a download.
    """
    # Get any currently downloaded resources
    current = self.history.find(resource.id, default={})

    # Make sure we haven't already downloaded the resource
    if current.get("version") == version:
      message = "'{}' already downloaded and up-to-date.".format(resource.id)
      self.messenger.send("update", message)

      return False

    return True



#!/usr/bin/env python
"""NCBI Human genome assembly."""

from ..resource import BaseResource
from ..servers.ncbi import NCBI
import sh


class Resource(BaseResource):
  """docstring for CCDS Resource"""
  def __init__(self):
    super(Resource, self).__init__()

    self.id = "ncbi_assembly"

    self.ftp = NCBI()
    self.baseUrl = "genomes/Homo_sapiens"

    self.parts = 25  # Including MT
    self.names = ["NCBI.Homo_sapiens.{chrom}.fa.gz".format(chrom=chrom)
                  for chrom in range(1, 23) + ["X", "Y", "MT"]]

  def versions(self):
    # Basically the combination of assembly + path is a valid float

    # First get versions from the archive (<=37.3)
    # Assemlies earlier than 35 are organized differently...
    archive = [dirName.replace("BUILD.", "")
               for dirName in self.ftp.ls(self.baseUrl + "/ARCHIVE")
               if dirName.startswith("BUILD")
               and not dirName.startswith("BUILD.34")]

    return archive + ["latest"]

  def newer(self, current, challenger):
    if current == "latest":
      return False  # Always
    elif challenger == "latest":
      return True  # Always
    else:
      # For now...
      return float(challenger) > float(current)

  def latest(self):
    # Return highest version number (float)
    return "latest"

  def paths(self, version):
    if version == "latest":
      base = ("{base}/Assembled_chromosomes/seq".format(base=self.baseUrl))

    else:
      # For all the archived releases
      base = ("{base}/ARCHIVE/BUILD.{v}/Assembled_chromosomes/seq"
              .format(base=self.baseUrl, v=version))

    files = self.ftp.listFiles(base, "hs_ref_*_chr*.fa.gz")

    # Should only be one file that matches
    return ["{base}/{file}".format(base=base, file=f) for f in files]

  def postClone(self, cloned_files, target_dir, version):
    """
    .. versionadded:: 0.3.0
    """
    # Start by extracting all the files
    for f in cloned_files:
      # GunZIP the file (and remove the archive)
      sh.gunzip(f)

    # Then let's concat them
    target_path = "{}/NCBI.Homo_sapiens.fa".format(target_dir)
    # Remove ".gz" ending to point to extracted files
    cat_args = [f[:-3] for f in cloned_files]

    # Execute the concatenation in the background and write to the target path
    sh.cat(*cat_args, _out=target_path, _bg=True)



import contextlib
import os
import importlib


@contextlib.contextmanager
def cd(path):
  """
  A context manager which changes the working directory to the given
  path, and then changes it back to its previous value on exit.
  """
  prev_cwd = os.getcwd()
  os.chdir(path)
  try:
    yield
  finally:
    os.chdir(prev_cwd)

def memoize(f):
  """
  Memoization decorator for a function taking one or more arguments.
  """
  class memodict(dict):
    def __getitem__(self, *key):
      return dict.__getitem__(self, key)

    def __missing__(self, key):
      ret = self[key] = f(*key)
      return ret

  return memodict().__getitem__

def load_class(full_class_string):
  """
  Dynamically load a class from a string.
  """
  class_data = full_class_string.split(".")
  module_path = ".".join(class_data[:-1])
  class_str = class_data[-1]

  module = importlib.import_module(module_path)
  # Finally, we retrieve the Class
  return getattr(module, class_str)



#!/usr/bin/env python
"""
Cosmid - A genomics resource manager
=====================================
There's a wealth of publicly available genomics resources; assemblies, database dumps, sample data etc. Problem is, they are spread across various FTP servers all over the world. Cosmid can help you find, clone and manage resources, including a quick way to keep up-to-date on the latest releases.

::

  # Clone dependencies from './cosmid.yaml'
  $ cosmid clone
  # Clone a specific resource + target verion (optional)
  $ cosmid clone <package>#<version>

  # Walk through setting up a new 'cosmid.yaml' file
  $ cosmid init
"""

# Bootstrap distribute unless already installed
from cosmid.distribute_setup import use_setuptools
use_setuptools()

import os
import sys

from setuptools import setup, find_packages

import cosmid

# Shortcut for publishing to Pypi
# Source: https://github.com/kennethreitz/tablib/blob/develop/setup.py
if sys.argv[-1] == "publish":
  os.system("python setup.py sdist upload")
  sys.exit()

setup(
  name = "cosmid",
  version = cosmid.__version__,
  packages = find_packages(exclude=["tests"]),
  scripts = ["scripts/cosmid"],

  # Project dependencies
  install_requires = [
    "docopt",
    "path.py",
    "pyyaml",
    "fuzzywuzzy",
    "termcolor",
    "sh"
  ],

  # Packages required for testing
  tests_require = [
    "nose"
  ],

  # Metadate for upload to Pypi
  author = "Robin Andeer",
  author_email = "robin.andeer@scilifelab.se",
  description = "A genomics resource manager",
  long_description = __doc__,
  license = "MIT",
  keywords = "bioinformatics resource manager database genomics",
  url = "https://github.com/robinandeer/cosmid",
  download_url = "https://github.com/robinandeer/cosmid/releases",
  classifiers = (
    "Development Status :: 3 - Alpha",
    "Natural Language :: English",
    "License :: OSI Approved :: MIT License",
    "Programming Language :: Python"
  )
)



#!/usr/bin/env python

from __future__ import print_function
from termcolor import colored


class Messenger(object):
  """
  Messenger Hub for sending formatted messages to the console. A number of
  pre-defined categories exist like "warning", "error", "update" as shortcuts.

  :param str sender: The program handle that will send the messages.
  """
  def __init__(self, sender=None):
    super(Messenger, self).__init__()

    # Who is sending the message?
    self.sender = sender

  def welcome(self, version):
    """
    <public> Returns a colored welcome messages (ascii art) ready to print.

    :param str version: A semver version sans leading 'v'.
    :returns: The colored welcome message
    """
    v = "=================| version {} |================".format(version)

    ascii = """
     .o88b.  .d88b.  .d8888. .88b  d88. d888888b d8888b.
    d8P  Y8 .8P  Y8. 88'  YP 88'YbdP`88   `88'   88  `8D
    8P      88    88 `8bo.   88  88  88    88    88   88
    8b      88    88   `Y8b. 88  88  88    88    88   88
    Y8b  d8 `8b  d8' db   8D 88  88  88   .88.   88  .8D
     `Y88P'  `Y88P'  `8888Y' YP  YP  YP Y888888P Y8888D'

     {}
    """.format(colored(v, "white"))

    return colored(ascii, "cyan")

  def send(self, category, message):
    """
    <public> Send a message to the console formatted according to the category.
    Different categories are formatted with different colors. The sender ID
    will be printed first.

    :param str message: The actual message to output to the console
    :param str category: (optional) The type of message (see below)
    :returns: self
    """
    # Pythonic switch statement
    statement = {
      "warning": self._warn(),
      "error": self._error(),
      "update": self._update(),
      "ghost": self._ghost()
    }.get(category, self._note())

    # Print the parts with tab-separation
    print("\t".join((self.sender, statement, message)))

    return self

  def _warn(self):
    """
    <private> Formats a yellow warning message.

    :returns: Colored string
    :rtype: str
    """
    return colored("WARN", "yellow", attrs=["bold"])

  def _error(self):
    """
    <private> Formats a red error message.

    :returns: Colored string
    :rtype: str
    """
    return colored("ERROR", "red", attrs=["bold"])

  def _note(self):
    """
    <private> Formats a white note.

    :returns: Colored string
    :rtype: str
    """
    return colored("NOTE", "white", attrs=["bold"])

  def _update(self):
    """
    <private> Formats a green update message.

    :returns: Colored string
    :rtype: str
    """
    return colored("UPDATE", "green", attrs=["bold"])

  def _ghost(self):
    """
    <private> Formats a transparent white note.

    :returns: Colored string
    :rtype: str
    """
    return colored("NOTE", "white", attrs=["bold", "dark"])



#!/usr/bin/env python
"""A few example BAM/Fasta files for testing provided by GATK."""

from ..resource import BaseResource
from ..servers.gatk import GATK
import sh


class Resource(BaseResource):
  """docstring for exampleBAM Resource"""
  def __init__(self):
    super(Resource, self).__init__()

    self.id = "example"

    self.ftp = GATK()
    self.baseUrl = "bundle"

    self.parts = 4
    self.names = ["exampleBAM.bam", "exampleBAM.bam.bai.gz",
                  "exampleFASTA.fasta.fai.gz", "exampleFASTA.fasta.gz"]

  def versions(self):
    return [float(dirName) for dirName in self.ftp.ls(self.baseUrl)]

  def defineVersion(version_tag, assembly='b37'):
    """
    <public> Determines the tuple representing the 'version' of a GATK bundle
    resource.

    .. versionadded:: 0.4.7

    :param str version_tag: The version string input by user
    :param str assembly: (optional) The default assembly to use if not
                         provided by user
    :returns: Tuple with bundle_id and assembly
    """
    version_parts = version_tag.split('/')
    bundle_id = version_parts[0]

    if '/' in version_tag:
      assembly = version_parts[1]

    return bundle_id, assembly

  def latest(self):
    return max(self.versions())

  def newer(self, current, challenger):
    # Simply check which float is biggest
    return float(challenger) > float(current)

  def paths(self, version):
    bundle_id, assembly = self.defineVersion(version)

    # 4 files
    base = "{base}/{v}/exampleFASTA".format(base=self.baseUrl, v=bundle_id)

    return ["{base}/{file}".format(base=base, file=f) for f in self.names]

  def postClone(self, cloned_files, target_dir, version):
    """
    Extracts the compressed archives.

    .. versionadded:: 0.3.0
    """
    # GZIP the files (and remove the archive)
    for f in cloned_files:

      # Only some of the files needs to be gzipped
      if f.endswith(".gz"):
        sh.gunzip(f)

    return 0



#!/usr/bin/env python
from ..core import FTP


class ThousandG(FTP):
  """docstring for EBI 1000 Genomes FTP"""
  def __init__(self):
    super(ThousandG, self).__init__("ftp-trace.ncbi.nih.gov", "anonymous", "")



from __future__ import print_function

__title__ = "cosmid"
__version__ = "0.4.8"
__author__ = "Robin Andeer"
__license__ = "MIT"
__copyright__ = "MIT Copyright (c) 2013 Robin Andeer"
__docformat__ = "restructuredtext"
